from __future__ import annotations

from abc import ABC, abstractmethod

from typing import Any, Dict

from .Schema import Schema


class SchemaType(Schema, ABC):
    """
    Abstract base class for any schema with the "type" keyword
    """

    @property
    @abstractmethod
    def type_name(self) -> str:
        """
        :return: Value for the JSON schema "type" keyword
        """
        pass

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {"type": self.type_name}





from __future__ import annotations

from abc import abstractmethod

from collections.abc import MutableSequence

from typing import (
    Generic, TypeVar, Union, Optional, Any, Iterable,
    Dict, List, Iterator, overload
)

from ..base_schema import Schema
from ..exceptions import ValidationError

from .DOMElement import DOMElement
from . import DOMInfo
from .functions import document

T_co = TypeVar('T_co')


class DOMList(DOMElement, MutableSequence, Generic[T_co]):
    """
    An array element (corresponding to a Python list).
    """

    __json_element_data__: List[DOMElement] = None

    def __init__(
            self,
            value: Iterable,
            json_dom_info: Optional[DOMInfo] = None,
            item_type: Optional[Schema] = None
    ) -> None:
        """
        :param value:         A list (or any :class:`Typing.Iterable`) containing the data to populate this
                              object's items.
        :param json_dom_info: A :class:`~wysdom.dom.DOMInfo` named tuple containing information about this object's
                              position in the DOM.
        :param item_type:     A :class:`~wysdom.Schema` object specifying what constitutes a valid item in this array.
        """
        if value and not isinstance(value, Iterable):
            raise ValidationError(
                f"Cannot validate input. Object is not iterable: {value}"
            )
        super().__init__(None, json_dom_info)
        self.__json_element_data__ = []
        if item_type is not None:
            self.item_type = item_type
        self[:] = value

    @overload
    @abstractmethod
    def __getitem__(self, i: int) -> T_co: ...

    @overload
    @abstractmethod
    def __getitem__(self, s: slice) -> MutableSequence[T_co]: ...

    def __getitem__(self, i: Union[int, slice]) -> Union[T_co, MutableSequence[T_co]]:
        return self.__json_element_data__[i]

    @overload
    @abstractmethod
    def __setitem__(self, i: int, o: T_co) -> None: ...

    @overload
    @abstractmethod
    def __setitem__(self, s: slice, o: Iterable[T_co]) -> None: ...

    def __setitem__(self, i: Union[int, slice], o: Union[T_co, MutableSequence[T_co]]) -> None:
        if type(i) is int:
            self.__json_element_data__[i] = self._new_child_item(o)
        else:
            self.__json_element_data__[i] = (self._new_child_item(x) for x in o)

    def _new_child_item(self, item: Any) -> DOMElement:
        return self.item_type(
            item,
            DOMInfo(
                document=document(self),
                parent=self
            )
        )

    @overload
    @abstractmethod
    def __delitem__(self, i: int) -> None: ...

    @overload
    @abstractmethod
    def __delitem__(self, i: slice) -> None: ...

    def __delitem__(self, i: int) -> None:
        del self.__json_element_data__[i]

    def __len__(self) -> int:
        return len(self.__json_element_data__)

    def __repr__(self):
        return repr(self.__json_element_data__)

    def __str__(self):
        return str(self.__json_element_data__)

    def insert(self, index: int, item: Any) -> None:
        self.__json_element_data__.insert(index, self._new_child_item(item))

    def to_builtin(self) -> List[Any]:
        """
        Returns the contents of this DOM object as a Python builtin.

        :return: A Python list containing this object's data
        """
        return [
            (
                v.to_builtin()
                if isinstance(v, DOMElement)
                else v
            )
            for v in self
        ]

    def walk_elements(self) -> Iterator[DOMInfo]:
        yield self.__json_dom_info__
        for value in self:
            if isinstance(value, DOMElement):
                yield from value.walk_elements()
            else:
                yield DOMInfo(value, document(self), self, None)

    def __copy__(self) -> DOMList:
        cls = self.__class__
        return cls(list(self))

    def __deepcopy__(self, memo: Dict[int, DOMElement]) -> DOMList:
        cls = self.__class__
        result = cls(
            value=self.to_builtin(),
            json_dom_info=self.__json_dom_info__,
            _item_type=self.item_type
        )
        memo[id(self)] = result
        return result



from typing import Dict, Union, Set

from ..base_schema import Schema


# TODO: DRY with object_schema.SchemaObject?

class DOMProperties(object):
    """
    A container for property information for a :class:`.DOMObject`.
    """

    properties: Dict[str, Schema] = None
    required: Set[str] = None
    additional_properties: Union[bool, Schema] = False

    def __init__(
            self,
            properties: Dict[str, Schema] = None,
            required: Set[str] = None,
            additional_properties: Union[bool, Schema] = False
    ) -> None:
        """
        :param properties:            A dictionary of :class:`~wysdom.base_schema.Schema` objects
                                      defining the expected names and types of a :class:`.DOMObject`'s
                                      properties.
        :param required:              A set of the property names that are required for an instance to be valid.
        :param additional_properties: Defines whether a :class:`.DOMObject` permits additional
                                      dynamically-named properties. Can be True or False, or
                                      can be set to a specific :class:`~wysdom.Schema` to restrict the permitted
                                      types of any additional properties.
        """
        self.properties = properties or {}
        self.required = required or set()
        self.additional_properties = additional_properties



from typing import Any, Dict, Tuple, Iterable, Optional

from ..dom import DOMInfo
from ..exceptions import ValidationError
from ..base_schema import Schema


class SchemaAnyOf(Schema):
    """
    A schema requiring a match with any of the permitted schemas supplied.

    :param allowed_schemas: A list (or other Iterable) containing the permitted
                            `Schema` objects.
    :param schema_ref_name: An optional unique reference name to use when this schema
                            is referred to by other schemas.
    """

    _allowed_schemas: Tuple[Schema] = None
    schema_ref_name: Optional[str] = None

    def __init__(
            self,
            allowed_schemas: Iterable[Schema],
            schema_ref_name: Optional[str] = None
    ) -> None:
        self._allowed_schemas = tuple(allowed_schemas)
        self.schema_ref_name = schema_ref_name

    def __call__(
            self,
            value: Any,
            dom_info: DOMInfo = None
    ) -> Any:
        valid_schemas = [
            allowed_schema
            for allowed_schema in self.allowed_schemas
            if allowed_schema.is_valid(value)
        ]
        if len(valid_schemas) > 1:
            raise ValidationError(
                "Ambiguous validation, more than one schema "
                f"is valid: {valid_schemas}"
            )
        if len(valid_schemas) == 0:
            raise ValidationError(
                f"No valid schema was found for the supplied value: {value}"
            )
        return valid_schemas[0](value, dom_info)

    @property
    def allowed_schemas(self) -> Tuple[Schema]:
        return self._allowed_schemas

    @property
    def referenced_schemas(self) -> Dict[str, Schema]:
        referenced_schemas = {}
        for allowed_schema in self.allowed_schemas:
            referenced_schemas.update(allowed_schema.referenced_schemas)
        if isinstance(self.schema_ref_name, str):
            referenced_schemas[self.schema_ref_name] = self
        return referenced_schemas

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {
            'anyOf': [
                allowed_schema.jsonschema_ref_schema
                for allowed_schema in self.allowed_schemas
            ]
        }



from typing import Optional

from ..base_schema import Schema
from .DOMElement import DOMElement
from . import DOMInfo


def dom(element: DOMElement) -> DOMInfo:
    """
    Retrieve a :class:`.DOMInfo` object for a :class:`.DOMElement` containing information
    about that element's position in the DOM.

    :param element: A DOM element
    :return:        The :class:`.DOMInfo` object for that DOM element
    """
    return element.__json_dom_info__


def document(element: DOMElement) -> Optional[DOMElement]:
    """
    Retrieve the owning document for a :class:`.DOMElement`, if it exists.

    :param element: A DOM element
    :return:        The owning document for that DOM element, or None if none exists
    """
    return dom(element).document


def parent(element: DOMElement) -> Optional[DOMElement]:
    """
    Retrieve the parent element of a :class:`.DOMElement`, if it exists.

    :param element: A DOM element
    :return:        The parent element of that DOM element, or None of none exists
    """
    return dom(element).parent


def key(element: DOMElement) -> Optional[str]:
    """
    Retrieve the key of a particular :class:`.DOMElement` in its parent element, if it can be
    referred to by a key (i.e. if it its parent element is a :class:`collections.abc.Mapping`).

    :param element: A DOM element
    :return:        The key of that DOM element in its parent, or None if it has no key
    """
    return dom(element).element_key


def schema(element: DOMElement) -> Schema:
    """
    Retrieve the :class:`~wysdom.base_schema.Schema` object for
    a particular :class:`.DOMElement`.

    :param element: A DOM element
    :return:        The :class:`~wysdom.base_schema.Schema` object associated
                    with that DOM element
    """
    return element.__json_schema__()
